@@ -13,7 +13,6 @@
 #include "criteria.h"
 #include "types.h"
 
-
 static int32_t max(int32_t a, int32_t b) {
 	return (a > b) ? a : b;
 }
@@ -45,26 +44,20 @@ void init_default_config(struct mako_config *config) {
 	new_criteria->style.spec.invisible = true;
 	new_criteria->raw_string = strdup("(default group-index=0)");
 
+	// Define the default format for the hidden placeholder notification.
+	new_criteria = create_criteria(config);
+	init_empty_style(&new_criteria->style);
+	new_criteria->hidden = true;
+	new_criteria->spec.hidden = true;
+	new_criteria->style.format = strdup("(%h more)");
+	new_criteria->style.spec.format = true;
+	new_criteria->raw_string = strdup("(default hidden)");
+
 	init_empty_style(&config->superstyle);
 
-	init_empty_style(&config->hidden_style);
-	config->hidden_style.format = strdup("(%h more)");
-	config->hidden_style.spec.format = true;
-
-	config->output = strdup("");
-	config->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
-
-	config->max_visible = 5;
+	config->max_history = 5;
 	config->sort_criteria = MAKO_SORT_CRITERIA_TIME;
 	config->sort_asc = 0;
-
-	config->button_bindings.left = MAKO_BINDING_INVOKE_DEFAULT_ACTION;
-	config->button_bindings.right = MAKO_BINDING_DISMISS;
-	config->button_bindings.middle = MAKO_BINDING_NONE;
-	config->touch = MAKO_BINDING_DISMISS;
-
-	config->anchor =
-		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP | ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
 }
 
 void finish_config(struct mako_config *config) {
@@ -74,8 +67,6 @@ void finish_config(struct mako_config *config) {
 	}
 
 	finish_style(&config->superstyle);
-	finish_style(&config->hidden_style);
-	free(config->output);
 }
 
 void init_default_style(struct mako_style *style) {
@@ -106,6 +97,7 @@ void init_default_style(struct mako_style *style) {
 	style->font = strdup("monospace 10");
 	style->markup = true;
 	style->format = strdup("<b>%s</b>\n%b");
+	style->text_alignment = PANGO_ALIGN_LEFT;
 
 	style->actions = true;
 	style->default_timeout = 0;
@@ -119,6 +111,20 @@ void init_default_style(struct mako_style *style) {
 
 	style->group_criteria_spec.none = true;
 	style->invisible = false;
+	style->history = true;
+	style->icon_location = MAKO_ICON_LOCATION_LEFT;
+
+	style->output = strdup("");
+	style->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
+	style->max_visible = 5;
+
+	style->anchor =
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP | ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+
+	style->button_bindings.left.action = MAKO_BINDING_INVOKE_DEFAULT_ACTION;
+	style->button_bindings.right.action = MAKO_BINDING_DISMISS;
+	style->button_bindings.middle.action = MAKO_BINDING_NONE;
+	style->touch_binding.action = MAKO_BINDING_DISMISS;
 
 	// Everything in the default config is explicitly specified.
 	memset(&style->spec, true, sizeof(struct mako_style_spec));
@@ -128,9 +134,30 @@ void init_empty_style(struct mako_style *style) {
 	memset(style, 0, sizeof(struct mako_style));
 }
 
+static void finish_binding(struct mako_binding *binding) {
+	free(binding->command);
+}
+
 void finish_style(struct mako_style *style) {
+	finish_binding(&style->button_bindings.left);
+	finish_binding(&style->button_bindings.middle);
+	finish_binding(&style->button_bindings.right);
+	finish_binding(&style->touch_binding);
+	finish_binding(&style->notify_binding);
+	free(style->icon_path);
 	free(style->font);
 	free(style->format);
+	free(style->output);
+}
+
+static void copy_binding(struct mako_binding *dst,
+		const struct mako_binding *src) {
+	finish_binding(dst);
+
+	*dst = *src;
+	if (src->command != NULL) {
+		dst->command = strdup(src->command);
+	}
 }
 
 // Update `target` with the values specified in `style`. If a failure occurs,
@@ -141,6 +168,7 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 	char *new_font = NULL;
 	char *new_format = NULL;
 	char *new_icon_path = NULL;
+	char *new_output = NULL;
 
 	if (style->spec.font) {
 		new_font = strdup(style->font);
@@ -169,6 +197,17 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 		}
 	}
 
+	if (style->spec.output) {
+		new_output = strdup(style->output);
+		if (new_output == NULL) {
+			free(new_format);
+			free(new_font);
+			free(new_icon_path);
+			fprintf(stderr, "allocation failed\n");
+			return false;
+		}
+	}
+
 	// Now on to actually setting things!
 
 	if (style->spec.width) {
@@ -229,6 +268,11 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 		target->spec.format = true;
 	}
 
+	if (style->spec.text_alignment) {
+		target->text_alignment = style->text_alignment;
+		target->spec.text_alignment = true;
+	}
+
 	if (style->spec.actions) {
 		target->actions = style->actions;
 		target->spec.actions = true;
@@ -274,11 +318,65 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 		target->spec.invisible = true;
 	}
 
+	if (style->spec.history) {
+		target->history = style->history;
+		target->spec.history = true;
+	}
+
+	if (style->spec.icon_location) {
+		target->icon_location = style->icon_location;
+		target->spec.icon_location = true;
+	}
+
 	if (style->border_radius) {
 		target->border_radius = style->border_radius;
 		target->spec.border_radius = true;
 	}
 
+	if (style->spec.output) {
+		free(target->output);
+		target->output = new_output;
+		target->spec.output = true;
+	}
+
+	if (style->spec.anchor) {
+		target->anchor = style->anchor;
+		target->spec.anchor = true;
+	}
+
+	if (style->spec.layer) {
+		target->layer = style->layer;
+		target->spec.layer = true;
+	}
+
+	if (style->spec.max_visible) {
+		target->max_visible = style->max_visible;
+		target->spec.max_visible = true;
+	}
+
+	if (style->spec.button_bindings.left) {
+		copy_binding(&target->button_bindings.left, &style->button_bindings.left);
+		target->spec.button_bindings.left = true;
+	}
+	if (style->spec.button_bindings.middle) {
+		copy_binding(&target->button_bindings.middle, &style->button_bindings.middle);
+		target->spec.button_bindings.middle = true;
+	}
+	if (style->spec.button_bindings.right) {
+		copy_binding(&target->button_bindings.right, &style->button_bindings.right);
+		target->spec.button_bindings.right = true;
+	}
+
+	if (style->spec.touch_binding) {
+		copy_binding(&target->touch_binding, &style->touch_binding);
+		target->spec.touch_binding = true;
+	}
+
+	if (style->spec.notify_binding) {
+		copy_binding(&target->notify_binding, &style->notify_binding);
+		target->spec.notify_binding = true;
+	}
+
 	return true;
 }
 
@@ -300,6 +398,7 @@ bool apply_superset_style(
 	target->spec.default_timeout = true;
 	target->spec.markup = true;
 	target->spec.actions = true;
+	target->spec.history = true;
 	target->spec.format = true;
 
 	free(target->format);
@@ -337,6 +436,7 @@ bool apply_superset_style(
 
 		target->markup |= style->markup;
 		target->actions |= style->actions;
+		target->history |= style->history;
 
 		// We do need to be safe about this one though.
 		if (style->spec.format) {
@@ -356,10 +456,10 @@ bool apply_superset_style(
 				memcpy(&current_specifier, format_pos, 2);
 				if (!strstr(target->format, current_specifier)) {
 					memcpy(target_format_pos, format_pos, 2);
+					target_format_pos += 2; // This needs to go to the next slot.
 				}
 
 				++format_pos; // Enough to move to the next match.
-				target_format_pos += 2; // This needs to go to the next slot.
 			}
 		}
 	}
@@ -369,26 +469,7 @@ bool apply_superset_style(
 
 static bool apply_config_option(struct mako_config *config, const char *name,
 		const char *value) {
-	if (strcmp(name, "max-visible") == 0) {
-		return parse_int(value, &config->max_visible);
-	} else if (strcmp(name, "output") == 0) {
-		free(config->output);
-		config->output = strdup(value);
-		return true;
-	} else if (strcmp(name, "layer") == 0) {
-		if (strcmp(value, "background") == 0) {
-			config->layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
-		} else if (strcmp(value, "bottom") == 0) {
-			config->layer = ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM;
-		} else if (strcmp(value, "top") == 0) {
-			config->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
-		} else if (strcmp(value, "overlay") == 0) {
-			config->layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
-		} else {
-			return false;
-		}
-		return true;
-	} else if (strcmp(name, "sort") == 0) {
+	if (strcmp(name, "sort") == 0) {
 		if (strcmp(value, "+priority") == 0) {
 			config->sort_criteria |= MAKO_SORT_CRITERIA_URGENCY;
 			config->sort_asc |= MAKO_SORT_CRITERIA_URGENCY;
@@ -405,34 +486,17 @@ static bool apply_config_option(struct mako_config *config, const char *name,
 			return false;
 		}
 		return true;
-	} else if (strcmp(name, "anchor") == 0) {
-		if (strcmp(value, "top-right") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
-		} else if (strcmp(value, "top-center") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP;
-		} else if (strcmp(value, "top-left") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT;
-		} else if (strcmp(value, "bottom-right") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
-		} else if (strcmp(value, "bottom-center") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM;
-		} else if (strcmp(value, "bottom-left") == 0) {
-			config->anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM |
-				ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT;
-		} else if (strcmp(value, "center") == 0) {
-			config->anchor = 0;
-		} else {
-			return false;
-		}
-		return true;
+	} else if (strcmp(name, "max-history") == 0) {
+		return parse_int(value, &config->max_history);
 	}
 
 	return false;
 }
 
+static bool has_prefix(const char *str, const char *prefix) {
+	return strncmp(str, prefix, strlen(prefix)) == 0;
+}
+
 static bool apply_style_option(struct mako_style *style, const char *name,
 		const char *value) {
 	struct mako_style_spec *spec = &style->spec;
@@ -446,9 +510,9 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 	} else if (strcmp(name, "text-color") == 0) {
 		return spec->colors.text = parse_color(value, &style->colors.text);
 	} else if (strcmp(name, "width") == 0) {
-		return spec->width = parse_int(value, &style->width);
+		return spec->width = parse_int_ge(value, &style->width, 1);
 	} else if (strcmp(name, "height") == 0) {
-		return spec->height = parse_int(value, &style->height);
+		return spec->height = parse_int_ge(value, &style->height, 1);
 	} else if (strcmp(name, "margin") == 0) {
 		return spec->margin = parse_directional(value, &style->margin);
 	} else if (strcmp(name, "padding") == 0) {
@@ -459,7 +523,7 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 		}
 		return spec->padding;
 	} else if (strcmp(name, "border-size") == 0) {
-		return spec->border_size = parse_int(value, &style->border_size);
+		return spec->border_size = parse_int_ge(value, &style->border_size, 0);
 	} else if (strcmp(name, "border-color") == 0) {
 		return spec->colors.border = parse_color(value, &style->colors.border);
 	} else if (strcmp(name, "progress-color") == 0) {
@@ -472,9 +536,22 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 		fprintf(stderr, "Icon support not built in, ignoring icons setting.\n");
 		return true;
 #endif
+	} else if (strcmp(name, "icon-location") == 0) {
+		if (!strcmp(value, "left")) {
+			style->icon_location = MAKO_ICON_LOCATION_LEFT;
+		} else if (!strcmp(value, "right")) {
+			style->icon_location = MAKO_ICON_LOCATION_RIGHT;
+		} else if (!strcmp(value, "top")) {
+			style->icon_location = MAKO_ICON_LOCATION_TOP;
+		} else if (!strcmp(value, "bottom")) {
+			style->icon_location = MAKO_ICON_LOCATION_BOTTOM;
+		} else {
+			return false;
+		}
+		return spec->icon_location = true;
 	} else if (strcmp(name, "max-icon-size") == 0) {
 		return spec->max_icon_size =
-			parse_int(value, &style->max_icon_size);
+			parse_int_ge(value, &style->max_icon_size, 1);
 	} else if (strcmp(name, "icon-path") == 0) {
 		free(style->icon_path);
 		return spec->icon_path = !!(style->icon_path = strdup(value));
@@ -485,9 +562,21 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 	} else if (strcmp(name, "format") == 0) {
 		free(style->format);
 		return spec->format = parse_format(value, &style->format);
+	} else if (strcmp(name, "text-alignment") == 0) {
+		if (strcmp(value, "left") == 0) {
+			style->text_alignment = PANGO_ALIGN_LEFT;
+		} else if (strcmp(value, "center") == 0) {
+			style->text_alignment = PANGO_ALIGN_CENTER;
+		} else if (strcmp(value, "right") == 0) {
+			style->text_alignment = PANGO_ALIGN_RIGHT;
+		} else {
+			return false;
+		}
+		style->spec.text_alignment = true;
+		return true;
 	} else if (strcmp(name, "default-timeout") == 0) {
 		return spec->default_timeout =
-			parse_int(value, &style->default_timeout);
+			parse_int_ge(value, &style->default_timeout, 0);
 	} else if (strcmp(name, "ignore-timeout") == 0) {
 		return spec->ignore_timeout =
 			parse_boolean(value, &style->ignore_timeout);
@@ -496,18 +585,89 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 			parse_criteria_spec(value, &style->group_criteria_spec);
 	} else if (strcmp(name, "invisible") == 0) {
 		return spec->invisible = parse_boolean(value, &style->invisible);
+	} else if (strcmp(name, "history") == 0) {
+		return spec->history = parse_boolean(value, &style->history);
 	} else if (strcmp(name, "border-radius") == 0) {
-		spec->border_radius = parse_int(value, &style->border_radius);
+		spec->border_radius = parse_int_ge(value, &style->border_radius, 0);
 		if (spec->border_radius && spec->padding) {
 			style->padding.left = max(style->border_radius, style->padding.left);
 			style->padding.right = max(style->border_radius, style->padding.right);
 		}
 		return spec->border_radius;
+	} else if (strcmp(name, "max-visible") == 0) {
+		return style->spec.max_visible = parse_int(value, &style->max_visible);
+	} else if (strcmp(name, "output") == 0) {
+		free(style->output);
+		style->output = strdup(value);
+		style->spec.output = true;
+		return true;
+	} else if (strcmp(name, "layer") == 0) {
+		if (strcmp(value, "background") == 0) {
+			style->layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
+		} else if (strcmp(value, "bottom") == 0) {
+			style->layer = ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM;
+		} else if (strcmp(value, "top") == 0) {
+			style->layer = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
+		} else if (strcmp(value, "overlay") == 0) {
+			style->layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
+		} else {
+			return false;
+		}
+		style->spec.layer = true;
+		return true;
+	} else if (strcmp(name, "anchor") == 0) {
+		return spec->anchor = parse_anchor(value, &style->anchor);
+	} else if (has_prefix(name, "on-")) {
+		struct mako_binding binding = {0};
+		if (strcmp(value, "none") == 0) {
+			binding.action = MAKO_BINDING_NONE;
+		} else if (strcmp(value, "dismiss") == 0) {
+			binding.action = MAKO_BINDING_DISMISS;
+		} else if (strcmp(value, "dismiss-all") == 0) {
+			binding.action = MAKO_BINDING_DISMISS_ALL;
+		} else if (strcmp(value, "dismiss-group") == 0) {
+			binding.action = MAKO_BINDING_DISMISS_GROUP;
+		} else if (strcmp(value, "invoke-default-action") == 0) {
+			binding.action = MAKO_BINDING_INVOKE_DEFAULT_ACTION;
+		} else if (has_prefix(value, "exec ")) {
+			binding.action = MAKO_BINDING_EXEC;
+			binding.command = strdup(value + strlen("exec "));
+		} else {
+			return false;
+		}
+
+		if (strcmp(name, "on-button-left") == 0) {
+			copy_binding(&style->button_bindings.left, &binding);
+			style->spec.button_bindings.left = true;
+		} else if (strcmp(name, "on-button-right") == 0) {
+			copy_binding(&style->button_bindings.right, &binding);
+			style->spec.button_bindings.right = true;
+		} else if (strcmp(name, "on-button-middle") == 0) {
+			copy_binding(&style->button_bindings.middle, &binding);
+			style->spec.button_bindings.middle = true;
+		} else if (strcmp(name, "on-touch") == 0) {
+			copy_binding(&style->touch_binding, &binding);
+			style->spec.touch_binding = true;
+		} else if (strcmp(name, "on-notify") == 0) {
+			copy_binding(&style->notify_binding, &binding);
+			style->spec.notify_binding = true;
+		} else {
+			return false;
+		}
+
+		return true;
 	}
 
 	return false;
 }
 
+bool apply_global_option(struct mako_config *config, const char *name,
+		const char *value) {
+	struct mako_criteria *global = global_criteria(config);
+	return apply_style_option(&global->style, name, value) ||
+		apply_config_option(config, name, value);
+}
+
 static bool file_exists(const char *path) {
 	return path && access(path, R_OK) != -1;
 }
@@ -557,7 +717,7 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 
 	FILE *f = fopen(path, "r");
 	if (!f) {
-		fprintf(stderr, "Unable to open %s for reading", path);
+		fprintf(stderr, "Unable to open %s for reading\n", path);
 		free(path);
 		return -1;
 	}
@@ -585,13 +745,17 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 		}
 
 		if (line[0] == '[' && line[strlen(line) - 1] == ']') {
+			// Since we hit the end of the previous criteria section, validate
+			// that it doesn't break any rules before moving on.
+			if (criteria != NULL && !validate_criteria(criteria)) {
+				fprintf(stderr, "Invalid configuration in criteria: [%s]\n",
+						criteria->raw_string);
+				ret = -1;
+				break;
+			}
+
 			free(section);
 			section = strndup(line + 1, strlen(line) - 2);
-			if (strcmp(section, "hidden") == 0) {
-				// Skip making a criteria for the hidden section.
-				criteria = NULL;
-				continue;
-			}
 			criteria = create_criteria(config);
 			if (!parse_criteria(section, criteria)) {
 				fprintf(stderr, "[%s:%d] Invalid criteria definition\n", base,
@@ -612,23 +776,13 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 		bool valid_option = false;
 		eq[0] = '\0';
 
-		struct mako_style *target_style;
-		if (section != NULL && strcmp(section, "hidden") == 0) {
-			// The hidden criteria is a lie, we store the associated style
-			// directly on the config because there's no "real" notification
-			// object to match against it later.
-			target_style = &config->hidden_style;
-		} else {
-			assert(criteria != NULL);
-			target_style = &criteria->style;
-		}
-
-		valid_option = apply_style_option(target_style, line, eq + 1);
+		valid_option = apply_style_option(&criteria->style, line, eq + 1);
 
 		if (!valid_option && section == NULL) {
 			valid_option = apply_config_option(config, line, eq + 1);
 		}
 
+
 		if (!valid_option) {
 			fprintf(stderr, "[%s:%d] Failed to parse option '%s'\n",
 				base, lineno, line);
@@ -637,6 +791,14 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 		}
 	}
 
+	// Validate the final criteria section since there was no opening bracket
+	// after it to do this in the loop.
+	if (ret != -1 && criteria != NULL && !validate_criteria(criteria)) {
+		fprintf(stderr, "Invalid configuration in criteria: [%s]\n",
+				criteria->raw_string);
+		ret = -1;
+	}
+
 	free(section);
 	free(line);
 	fclose(f);
@@ -660,12 +822,15 @@ int parse_config_arguments(struct mako_config *config, int argc, char **argv) {
 		{"border-radius", required_argument, 0, 0},
 		{"progress-color", required_argument, 0, 0},
 		{"icons", required_argument, 0, 0},
+		{"icon-location", required_argument, 0, 0},
 		{"icon-path", required_argument, 0, 0},
 		{"max-icon-size", required_argument, 0, 0},
 		{"markup", required_argument, 0, 0},
 		{"actions", required_argument, 0, 0},
 		{"format", required_argument, 0, 0},
 		{"max-visible", required_argument, 0, 0},
+		{"max-history", required_argument, 0, 0},
+		{"history", required_argument, 0, 0},
 		{"default-timeout", required_argument, 0, 0},
 		{"ignore-timeout", required_argument, 0, 0},
 		{"output", required_argument, 0, 0},
@@ -673,28 +838,38 @@ int parse_config_arguments(struct mako_config *config, int argc, char **argv) {
 		{"anchor", required_argument, 0, 0},
 		{"sort", required_argument, 0, 0},
 		{"group-by", required_argument, 0, 0},
+		{"on-button-left", required_argument, 0, 0},
+		{"on-button-right", required_argument, 0, 0},
+		{"on-button-middle", required_argument, 0, 0},
+		{"on-touch", required_argument, 0, 0},
 		{0},
 	};
 
-	struct mako_criteria *root_criteria =
-		wl_container_of(config->criteria.next, root_criteria, link);
-
 	optind = 1;
 	char *config_arg = NULL;
+	int opt_status = 0;
 	while (1) {
 		int option_index = -1;
 		int c = getopt_long(argc, argv, "hc:", long_options, &option_index);
 		if (c < 0) {
 			break;
 		} else if (c == 'h') {
-			return 1;
+			opt_status = 1;
+			break;
 		} else if (c == 'c') {
+			free(config_arg);
 			config_arg = strdup(optarg);
 		} else if (c != 0) {
-			return -1;
+			opt_status = -1;
+			break;
 		}
 	}
 
+	if (opt_status != 0) {
+		free(config_arg);
+		return opt_status;
+	}
+
 	int config_status = load_config_file(config, config_arg);
 	if (config_status < 0) {
 		return -1;
@@ -713,8 +888,7 @@ int parse_config_arguments(struct mako_config *config, int argc, char **argv) {
 		}
 
 		const char *name = long_options[option_index].name;
-		if (!apply_style_option(&root_criteria->style, name, optarg)
-				&& !apply_config_option(config, name, optarg)) {
+		if (!apply_global_option(config, name, optarg)) {
 			fprintf(stderr, "Failed to parse option '%s'\n", name);
 			return -1;
 		}
