@@ -4,14 +4,21 @@
 #include <stdbool.h>
 #include <stdint.h>
 #include <wayland-client.h>
+#include <pango/pango.h>
 #include "types.h"
-#include "wlr-layer-shell-unstable-v1-client-protocol.h"
 
-enum mako_binding {
+enum mako_binding_action {
 	MAKO_BINDING_NONE,
 	MAKO_BINDING_DISMISS,
+	MAKO_BINDING_DISMISS_GROUP,
 	MAKO_BINDING_DISMISS_ALL,
 	MAKO_BINDING_INVOKE_DEFAULT_ACTION,
+	MAKO_BINDING_EXEC,
+};
+
+struct mako_binding {
+	enum mako_binding_action action;
+	char *command; // for MAKO_BINDING_EXEC
 };
 
 enum mako_sort_criteria {
@@ -19,17 +26,28 @@ enum mako_sort_criteria {
 	MAKO_SORT_CRITERIA_URGENCY = 2,
 };
 
+enum mako_icon_location {
+	MAKO_ICON_LOCATION_LEFT,
+	MAKO_ICON_LOCATION_RIGHT,
+	MAKO_ICON_LOCATION_TOP,
+	MAKO_ICON_LOCATION_BOTTOM,
+};
+
 // Represents which fields in the style were specified in this style. All
 // fields in the mako_style structure should have a counterpart here. Inline
 // structs are also mirrored.
 struct mako_style_spec {
 	bool width, height, margin, padding, border_size, border_radius, font,
-		markup, format, actions, default_timeout, ignore_timeout, icons,
-		max_icon_size, icon_path, group_criteria_spec, invisible;
-
+		markup, format, text_alignment, actions, default_timeout, ignore_timeout,
+		icons, max_icon_size, icon_path, group_criteria_spec, invisible, history,
+		icon_location, max_visible, layer, output, anchor;
 	struct {
 		bool background, text, border, progress;
 	} colors;
+	struct {
+		bool left, right, middle;
+	} button_bindings;
+	bool touch_binding, notify_binding;
 };
 
 
@@ -50,6 +68,7 @@ struct mako_style {
 	char *font;
 	bool markup;
 	char *format;
+	PangoAlignment text_alignment;
 
 	bool actions;
 	int default_timeout; // in ms
@@ -65,26 +84,28 @@ struct mako_style {
 	struct mako_criteria_spec group_criteria_spec;
 
 	bool invisible; // Skipped during render, doesn't count toward max_visible
-};
-
-struct mako_config {
-	struct wl_list criteria; // mako_criteria::link
+	bool history;
+	enum mako_icon_location icon_location;
 
 	int32_t max_visible;
 	char *output;
 	enum zwlr_layer_shell_v1_layer layer;
 	uint32_t anchor;
-	uint32_t sort_criteria; //enum mako_sort_criteria
-	uint32_t sort_asc;
-
-	struct mako_style hidden_style;
-	struct mako_style superstyle;
 
 	struct {
-		enum mako_binding left, right, middle;
+		struct mako_binding left, right, middle;
 	} button_bindings;
+	struct mako_binding touch_binding, notify_binding;
+};
 
-	enum mako_binding touch;
+struct mako_config {
+	struct wl_list criteria; // mako_criteria::link
+
+	uint32_t sort_criteria; //enum mako_sort_criteria
+	uint32_t sort_asc;
+	int32_t max_history;
+
+	struct mako_style superstyle;
 };
 
 void init_default_config(struct mako_config *config);
@@ -100,5 +121,7 @@ bool apply_superset_style(
 int parse_config_arguments(struct mako_config *config, int argc, char **argv);
 int load_config_file(struct mako_config *config, char *config_arg);
 int reload_config(struct mako_config *config, int argc, char **argv);
+bool apply_global_option(struct mako_config *config, const char *name,
+	const char *value);
 
 #endif
