@@ -1,6 +1,7 @@
 #ifndef MAKO_CRITERIA_H
 #define MAKO_CRITERIA_H
 
+#include <regex.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <wayland-client.h>
@@ -28,9 +29,18 @@ struct mako_criteria {
 	char *category;
 	char *desktop_entry;
 	char *summary;
+	regex_t summary_pattern;
 	char *body;
+	regex_t body_pattern;
+
+	char *mode;
+
+	// Second-pass matches:
 	int group_index;
 	bool grouped;  // Whether group_index is non-zero
+	char *output;
+	uint32_t anchor;
+	bool hidden;
 };
 
 struct mako_criteria *create_criteria(struct mako_config *config);
@@ -48,4 +58,6 @@ ssize_t apply_each_criteria(struct wl_list *criteria_list,
 struct mako_criteria *create_criteria_from_notification(
 		struct mako_notification *notif, struct mako_criteria_spec *spec);
 
+bool validate_criteria(struct mako_criteria *criteria);
+
 #endif
