@@ -2,10 +2,12 @@
 #define MAKO_DBUS_H
 
 #include <stdbool.h>
-#ifdef HAVE_SYSTEMD
+#ifdef HAVE_LIBSYSTEMD
 #include <systemd/sd-bus.h>
-#elif HAVE_ELOGIND
+#elif HAVE_LIBELOGIND
 #include <elogind/sd-bus.h>
+#elif HAVE_BASU
+#include <basu/sd-bus.h>
 #endif
 
 struct mako_state;
@@ -18,7 +20,8 @@ void finish_dbus(struct mako_state *state);
 void notify_notification_closed(struct mako_notification *notif,
 	enum mako_notification_close_reason reason);
 
-void notify_action_invoked(struct mako_action *action);
+void notify_action_invoked(struct mako_action *action,
+	const char *activation_token);
 
 int init_dbus_xdg(struct mako_state *state);
 
