@@ -5,10 +5,12 @@
 #include <stdbool.h>
 #include <time.h>
 #include <wayland-client.h>
-#ifdef HAVE_SYSTEMD
+#ifdef HAVE_LIBSYSTEMD
 #include <systemd/sd-bus.h>
-#elif HAVE_ELOGIND
+#elif HAVE_LIBELOGIND
 #include <elogind/sd-bus.h>
+#elif HAVE_BASU
+#include <basu/sd-bus.h>
 #endif
 
 enum mako_event {
